#version 460 core

#ifndef MATH_GLSL_GUARD
#define MATH_GLSL_GUARD

// Define max/min single precision float constants
#define FLT_MAX 3.402823466e+38
#define FLT_MIN 1.175494351e-38

// Define math constants
#define M_PI  3.14159265358979f
#define M_EPS 0.0000000001f

// Rounded-up division functions

#define CEIL_DIV(a, b) ((a + (b - 1)) / b)

int ceil_div(in int a, in int b) {return CEIL_DIV(a, b); }
uint ceil_div(in uint a, in uint b) {return CEIL_DIV(a, b); }

// Horizontal component product functions

#define HPROD_2(v) (v.x * v.y)
#define HPROD_3(v) (v.x * v.y * v.z)
#define HPROD_4(v) (v.x * v.y * v.z * v.w)

uint  hprod(in uvec2 v) { return HPROD_2(v); }
uint  hprod(in uvec3 v) { return HPROD_3(v); }
uint  hprod(in uvec4 v) { return HPROD_4(v); }
int   hprod(in ivec2 v) { return HPROD_2(v); }
int   hprod(in ivec3 v) { return HPROD_3(v); }
int   hprod(in ivec4 v) { return HPROD_4(v); }
float hprod(in vec2 v)  { return HPROD_2(v); }
float hprod(in vec3 v)  { return HPROD_3(v); }
float hprod(in vec4 v)  { return HPROD_4(v); }

// Horizontal component sum functions

#define HSUM_2(v) (v.x + v.y)
#define HSUM_3(v) (v.x + v.y + v.z)
#define HSUM_4(v) (v.x + v.y + v.z + v.w)

uint   hsum(in uvec2 v) { return HSUM_2(v); }
uint   hsum(in uvec3 v) { return HSUM_3(v); }
uint   hsum(in uvec4 v) { return HSUM_4(v); }
int    hsum(in ivec2 v) { return HSUM_2(v); }
int    hsum(in ivec3 v) { return HSUM_3(v); }
int    hsum(in ivec4 v) { return HSUM_4(v); }
float  hsum(in vec2 v)  { return HSUM_2(v); }
float  hsum(in vec3 v)  { return HSUM_3(v); }
float  hsum(in vec4 v)  { return HSUM_4(v); }
double hsum(in dvec2 v) { return HSUM_2(v); }
double hsum(in dvec3 v) { return HSUM_3(v); }
double hsum(in dvec4 v) { return HSUM_4(v); }

// Horizontal component min functions

#define HMIN_2(v) min(v.x, v.y)
#define HMIN_3(v) min(min(v.x, v.y), v.z)
#define HMIN_4(v) min(min(min(v.x, v.y), v.z), v.w)

uint  hmin(in uvec2 v) { return HMIN_2(v); }
uint  hmin(in uvec3 v) { return HMIN_3(v); }
uint  hmin(in uvec4 v) { return HMIN_4(v); }
int   hmin(in ivec2 v) { return HMIN_2(v); }
int   hmin(in ivec3 v) { return HMIN_3(v); }
int   hmin(in ivec4 v) { return HMIN_4(v); }
float hmin(in vec2 v)  { return HMIN_2(v); }
float hmin(in vec3 v)  { return HMIN_3(v); }
float hmin(in vec4 v)  { return HMIN_4(v); }

// Horizontal component max functions

#define HMAX_2(v) max(v.x, v.y)
#define HMAX_3(v) max(max(v.x, v.y), v.z)
#define HMAX_4(v) max(max(max(v.x, v.y), v.z), v.w)

uint  hmax(in uvec2 v) { return HMAX_2(v); }
uint  hmax(in uvec3 v) { return HMAX_3(v); }
uint  hmax(in uvec4 v) { return HMAX_4(v); }
int   hmax(in ivec2 v) { return HMAX_2(v); }
int   hmax(in ivec3 v) { return HMAX_3(v); }
int   hmax(in ivec4 v) { return HMAX_4(v); }
float hmax(in vec2 v)  { return HMAX_2(v); }
float hmax(in vec3 v)  { return HMAX_3(v); }
float hmax(in vec4 v)  { return HMAX_4(v); }

// Horizontal component self-dot functions

float sdot(in vec2 v) { return dot(v, v); }
float sdot(in vec3 v) { return dot(v, v); }
float sdot(in vec4 v) { return dot(v, v); }

#endif // MATH_GLSL_GUARD


// Tree node data structure
struct Node {
  vec3 minb;
  uint i;
  vec3 maxb;
  uint n;
};

// Layout declarations
layout(std140) uniform;
layout(std430) buffer;

const vec3 mult_data[8] = vec3[8](
  vec3(0, 0, 0), // 0
  vec3(0, 0, 1), // 1
  vec3(0, 1, 0), // 2
  vec3(0, 1, 1), // 3
  vec3(1, 0, 0), // 4
  vec3(1, 0, 1), // 5
  vec3(1, 1, 0), // 6
  vec3(1, 1, 1)  // 7
);

const uint minb_data [8]= uint[8](
  7, // 111 - 000 ; 0    ... 0
  3, // 011 - 100 ; 4    ... 1
  1, // 001 - 110 ; 6    ... 2
  5, // 101 - 010 ; 2    ... 3
  
  6, // 110 - 001 ; 1    ... 4
  2, // 010 - 101 ; 5    ... 5
  0, // 000 - 111 ; 7    ... 6
  4  // 100 - 011 ; 3    ... 7
);

const uint elem_data[48] = uint[48](
  0, 1, 1, 2, 2, 3, 3, 0,
  1, 5, 5, 6, 6, 2, 2, 1,
  0, 3, 3, 7, 7, 4, 4, 0,
  0, 4, 4, 5, 5, 1, 1, 0,
  3, 2, 2, 6, 6, 7, 7, 3,
  5, 4, 4, 7, 7, 6, 6, 5
);

// Buffer declarations
layout(binding = 0) restrict readonly buffer b_tree { Node data[]; } tree_in;
layout(binding = 0) uniform b_unif {
  uint nodi;
  uint nodn;
} unif;
layout(binding = 1) uniform b_camr {
  mat4 matrix;
  vec2 aspect;
} camera;

void main() {
  uint i = unif.nodi + gl_VertexID / 48, j = gl_VertexID % 48;

  uint mult_minb_i = minb_data[elem_data[j]];
  uint mult_maxb_i = 7 - mult_minb_i;
  
  Node node = tree_in.data[i];
  if (node.n == 0) {
    gl_Position = vec4(0);
  } else if (node.n == 1) {
    vec3 v = (node.minb - vec3(0.0025)) * mult_data[mult_minb_i] +
             (node.maxb + vec3(0.0025)) * mult_data[mult_maxb_i];
    gl_Position = camera.matrix * vec4(v, 1);
  } else {
    vec3 v = node.minb * mult_data[mult_minb_i] +
             node.maxb * mult_data[mult_maxb_i];
    gl_Position = camera.matrix * vec4(v, 1);
  }
}
