#version 460 core

#ifndef MATH_GLSL_GUARD
#define MATH_GLSL_GUARD

// Define max/min single precision float constants
#define FLT_MAX 3.402823466e+38
#define FLT_MIN 1.175494351e-38

// Define math constants
#define M_PI  3.14159265358979f
#define M_EPS 0.0000000001f

// Rounded-up division functions

#define CEIL_DIV(a, b) ((a + (b - 1)) / b)

int ceil_div(in int a, in int b) {return CEIL_DIV(a, b); }
uint ceil_div(in uint a, in uint b) {return CEIL_DIV(a, b); }

// Horizontal component product functions

#define HPROD_2(v) (v.x * v.y)
#define HPROD_3(v) (v.x * v.y * v.z)
#define HPROD_4(v) (v.x * v.y * v.z * v.w)

uint  hprod(in uvec2 v) { return HPROD_2(v); }
uint  hprod(in uvec3 v) { return HPROD_3(v); }
uint  hprod(in uvec4 v) { return HPROD_4(v); }
int   hprod(in ivec2 v) { return HPROD_2(v); }
int   hprod(in ivec3 v) { return HPROD_3(v); }
int   hprod(in ivec4 v) { return HPROD_4(v); }
float hprod(in vec2 v)  { return HPROD_2(v); }
float hprod(in vec3 v)  { return HPROD_3(v); }
float hprod(in vec4 v)  { return HPROD_4(v); }

// Horizontal component sum functions

#define HSUM_2(v) (v.x + v.y)
#define HSUM_3(v) (v.x + v.y + v.z)
#define HSUM_4(v) (v.x + v.y + v.z + v.w)

uint   hsum(in uvec2 v) { return HSUM_2(v); }
uint   hsum(in uvec3 v) { return HSUM_3(v); }
uint   hsum(in uvec4 v) { return HSUM_4(v); }
int    hsum(in ivec2 v) { return HSUM_2(v); }
int    hsum(in ivec3 v) { return HSUM_3(v); }
int    hsum(in ivec4 v) { return HSUM_4(v); }
float  hsum(in vec2 v)  { return HSUM_2(v); }
float  hsum(in vec3 v)  { return HSUM_3(v); }
float  hsum(in vec4 v)  { return HSUM_4(v); }
double hsum(in dvec2 v) { return HSUM_2(v); }
double hsum(in dvec3 v) { return HSUM_3(v); }
double hsum(in dvec4 v) { return HSUM_4(v); }

// Horizontal component min functions

#define HMIN_2(v) min(v.x, v.y)
#define HMIN_3(v) min(min(v.x, v.y), v.z)
#define HMIN_4(v) min(min(min(v.x, v.y), v.z), v.w)

uint  hmin(in uvec2 v) { return HMIN_2(v); }
uint  hmin(in uvec3 v) { return HMIN_3(v); }
uint  hmin(in uvec4 v) { return HMIN_4(v); }
int   hmin(in ivec2 v) { return HMIN_2(v); }
int   hmin(in ivec3 v) { return HMIN_3(v); }
int   hmin(in ivec4 v) { return HMIN_4(v); }
float hmin(in vec2 v)  { return HMIN_2(v); }
float hmin(in vec3 v)  { return HMIN_3(v); }
float hmin(in vec4 v)  { return HMIN_4(v); }

// Horizontal component max functions

#define HMAX_2(v) max(v.x, v.y)
#define HMAX_3(v) max(max(v.x, v.y), v.z)
#define HMAX_4(v) max(max(max(v.x, v.y), v.z), v.w)

uint  hmax(in uvec2 v) { return HMAX_2(v); }
uint  hmax(in uvec3 v) { return HMAX_3(v); }
uint  hmax(in uvec4 v) { return HMAX_4(v); }
int   hmax(in ivec2 v) { return HMAX_2(v); }
int   hmax(in ivec3 v) { return HMAX_3(v); }
int   hmax(in ivec4 v) { return HMAX_4(v); }
float hmax(in vec2 v)  { return HMAX_2(v); }
float hmax(in vec3 v)  { return HMAX_3(v); }
float hmax(in vec4 v)  { return HMAX_4(v); }

// Horizontal component self-dot functions

float sdot(in vec2 v) { return dot(v, v); }
float sdot(in vec3 v) { return dot(v, v); }
float sdot(in vec4 v) { return dot(v, v); }

#endif // MATH_GLSL_GUARD


// Fragment input declarations
layout(location = 0) in vec2 in_value_vert;

// Fragment output declarations
layout(location = 0) out vec4 out_value_colr;

// Buffer declarations
layout(binding = 0) restrict readonly buffer b_data { vec2 data[];  } data_in;
layout(binding = 1) restrict readonly buffer b_bary { vec4 data[];  } bary_in;
layout(binding = 2) restrict readonly buffer b_vert { vec3 data[];  } vert_in;
layout(binding = 3) restrict readonly buffer b_elem { uvec4 data[]; } elem_in;
layout(binding = 0) uniform b_unif {
  mat4  camera_matrix;
  uvec2 size_in;
  uint  n_verts;
  uint  n_quads;
} unif_in;

void main() {
  if (clamp(in_value_vert, 0, 1) != in_value_vert)
    discard;
  else {
    // Nearest-neighbor find corresponding 2d/1d coordinates in input size
    uvec2 ij_in = uvec2(in_value_vert * vec2(unif_in.size_in));
    uint i_in = unif_in.size_in.x * ij_in.y + ij_in.x;

    // Load relevant barycentric weights, vertex indices, constraint indices
    uint elems_offs = gl_PrimitiveID * unif_in.n_verts;
    uvec4 elems_idx = elem_in.data[floatBitsToUint(bary_in.data[i_in].w)];
    vec3 xyz  = bary_in.data[i_in].xyz;
    vec4 bary = vec4(xyz, 1.f - hsum(xyz));

    // Compute convex combination of vertex colors
    vec3 v = vec3(0);
    for (uint j = 0; j < 4; ++j)
      v += bary[j] * vert_in.data[elems_offs + elems_idx[j]];

    out_value_colr = vec4(v, 1);
  }
}
